@@ -23,6 +23,7 @@
 #include <config.h>
 #endif
 
+#include <errno.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
@@ -44,8 +45,9 @@ struct connman_route {
 };
 
 struct connman_provider {
-	struct connman_element element;
+	int refcount;
 	struct connman_service *vpn_service;
+	int index;
 	char *identifier;
 	char *name;
 	char *type;
@@ -55,6 +57,10 @@ struct connman_provider {
 	GHashTable *routes;
 	struct connman_provider_driver *driver;
 	void *driver_data;
+	GHashTable *setting_strings;
+	GHashTable *user_routes;
+	gchar **user_networks;
+	gsize num_user_networks;
 };
 
 void __connman_provider_append_properties(struct connman_provider *provider,
@@ -73,6 +79,174 @@ void __connman_provider_append_properties(struct connman_provider *provider,
 						 &provider->type);
 }
 
+int __connman_provider_append_user_route(struct connman_provider *provider,
+			int family, const char *network, const char *netmask)
+{
+	struct connman_route *route;
+	char *key = g_strdup_printf("%d/%s/%s", family, network, netmask);
+
+	DBG("family %d network %s netmask %s", family, network, netmask);
+
+	route = g_hash_table_lookup(provider->user_routes, key);
+	if (route == NULL) {
+		route = g_try_new0(struct connman_route, 1);
+		if (route == NULL) {
+			connman_error("out of memory");
+			return -ENOMEM;
+		}
+
+		route->family = family;
+		route->host = g_strdup(network);
+		route->netmask = g_strdup(netmask);
+
+		g_hash_table_replace(provider->user_routes, key, route);
+	} else
+		g_free(key);
+
+	return 0;
+}
+
+static void set_user_networks(struct connman_provider *provider,
+							char **networks)
+{
+	int i = 0;
+
+	while (networks[i] != NULL) {
+		char **elems = g_strsplit(networks[i], "/", 0);
+		char *network, *netmask = NULL;
+		int family = PF_UNSPEC, ret;
+
+		if (elems == NULL)
+			break;
+
+		network = elems[0];
+		if (elems[1] != NULL)
+			netmask = elems[1];
+
+		if (g_strrstr(network, ":") != NULL)
+			family = AF_INET6;
+		else if (g_strrstr(network, ".") != NULL) {
+			family = AF_INET;
+
+			if (g_strrstr(netmask, ".") == NULL) {
+				/* We have netmask length */
+				in_addr_t addr;
+				struct in_addr netmask_in;
+				unsigned char prefix_len = atoi(netmask);
+
+				addr = 0xffffffff << (32 - prefix_len);
+				netmask_in.s_addr = htonl(addr);
+				netmask = inet_ntoa(netmask_in);
+
+				DBG("network %s netmask %s", network, netmask);
+			}
+		}
+
+		ret = __connman_provider_append_user_route(provider,
+						family, network, netmask);
+		g_strfreev(elems);
+
+		if (ret != 0)
+			break;
+
+		i++;
+	}
+}
+
+static int provider_load_from_keyfile(struct connman_provider *provider,
+		GKeyFile *keyfile)
+{
+	gsize idx = 0;
+	gchar **settings;
+	gchar *key, *value;
+	gsize length;
+
+	settings = g_key_file_get_keys(keyfile, provider->identifier, &length,
+				NULL);
+	if (settings == NULL) {
+		g_key_file_free(keyfile);
+		return -ENOENT;
+	}
+
+	while (idx < length) {
+		key = settings[idx];
+		if (key != NULL) {
+			if (g_str_equal(key, "Networks") == TRUE) {
+				g_strfreev(provider->user_networks);
+				provider->user_networks =
+					g_key_file_get_string_list(keyfile,
+						provider->identifier,
+						key,
+						&provider->num_user_networks,
+						NULL);
+			} else {
+				value = g_key_file_get_string(keyfile,
+							provider->identifier,
+							key, NULL);
+				connman_provider_set_string(provider, key,
+							value);
+				g_free(value);
+			}
+		}
+		idx += 1;
+	}
+	g_strfreev(settings);
+
+	if (provider->user_networks != NULL)
+		set_user_networks(provider, provider->user_networks);
+
+	return 0;
+}
+
+static int connman_provider_load(struct connman_provider *provider)
+{
+	GKeyFile *keyfile;
+
+	DBG("provider %p", provider);
+
+	keyfile = __connman_storage_load_provider(provider->identifier);
+	if (keyfile == NULL)
+		return -ENOENT;
+
+	provider_load_from_keyfile(provider, keyfile);
+
+	g_key_file_free(keyfile);
+	return 0;
+}
+
+static int connman_provider_save(struct connman_provider *provider)
+{
+	GKeyFile *keyfile;
+
+	DBG("provider %p", provider);
+
+	keyfile = g_key_file_new();
+	if (keyfile == NULL)
+		return -ENOMEM;
+
+	g_key_file_set_string(keyfile, provider->identifier,
+			"Name", provider->name);
+	g_key_file_set_string(keyfile, provider->identifier,
+			"Type", provider->type);
+	g_key_file_set_string(keyfile, provider->identifier,
+			"Host", provider->host);
+	g_key_file_set_string(keyfile, provider->identifier,
+			"VPN.Domain", provider->domain);
+	if (provider->user_networks != NULL)
+		g_key_file_set_string_list(keyfile, provider->identifier,
+				"Networks",
+				(const gchar **)provider->user_networks,
+				provider->num_user_networks);
+
+	if (provider->driver != NULL && provider->driver->save != NULL)
+		provider->driver->save(provider, keyfile);
+
+	__connman_storage_save_provider(keyfile, provider->identifier);
+        g_key_file_free(keyfile);
+
+	return 0;
+}
+
 static struct connman_provider *connman_provider_lookup(const char *identifier)
 {
 	struct connman_provider *provider = NULL;
@@ -82,23 +256,6 @@ static struct connman_provider *connman_provider_lookup(const char *identifier)
 	return provider;
 }
 
-struct connman_provider *connman_provider_ref(struct connman_provider *provider)
-{
-	DBG("provider %p", provider);
-
-	if (connman_element_ref(&provider->element) == NULL)
-		return NULL;
-
-	return provider;
-}
-
-void connman_provider_unref(struct connman_provider *provider)
-{
-	DBG("provider %p", provider);
-
-	connman_element_unref(&provider->element);
-}
-
 static gboolean match_driver(struct connman_provider *provider,
 				struct connman_provider_driver *driver)
 {
@@ -137,16 +294,76 @@ static int provider_probe(struct connman_provider *provider)
 	return 0;
 }
 
+static void provider_remove(struct connman_provider *provider)
+{
+	if (provider->driver != NULL) {
+		provider->driver->remove(provider);
+		provider->driver = NULL;
+	}
+}
+
+static int provider_register(struct connman_provider *provider)
+{
+	return provider_probe(provider);
+}
+
+static void provider_unregister(struct connman_provider *provider)
+{
+	provider_remove(provider);
+}
+
+struct connman_provider *
+connman_provider_ref_debug(struct connman_provider *provider,
+			const char *file, int line, const char *caller)
+{
+	DBG("%p ref %d by %s:%d:%s()", provider, provider->refcount + 1,
+		file, line, caller);
+
+	__sync_fetch_and_add(&provider->refcount, 1);
+
+	return provider;
+}
+
+static void provider_destruct(struct connman_provider *provider)
+{
+	DBG("provider %p", provider);
+
+	g_free(provider->name);
+	g_free(provider->type);
+	g_free(provider->host);
+	g_free(provider->domain);
+	g_free(provider->identifier);
+	g_strfreev(provider->user_networks);
+	g_hash_table_destroy(provider->routes);
+	g_hash_table_destroy(provider->user_routes);
+	g_hash_table_destroy(provider->setting_strings);
+	g_free(provider);
+}
+
+void connman_provider_unref_debug(struct connman_provider *provider,
+				const char *file, int line, const char *caller)
+{
+	DBG("%p ref %d by %s:%d:%s()", provider, provider->refcount - 1,
+		file, line, caller);
+
+	if (__sync_fetch_and_sub(&provider->refcount, 1) != 1)
+		return;
+
+	provider_remove(provider);
+
+	provider_destruct(provider);
+}
+
 static int provider_indicate_state(struct connman_provider *provider,
 					enum connman_service_state state)
 {
 	DBG("state %d", state);
 
-	__connman_service_indicate_state(provider->vpn_service, state,
+	__connman_service_ipconfig_indicate_state(provider->vpn_service, state,
 					CONNMAN_IPCONFIG_TYPE_IPV4);
 
-	return __connman_service_indicate_state(provider->vpn_service, state,
-					CONNMAN_IPCONFIG_TYPE_IPV6);
+	return __connman_service_ipconfig_indicate_state(provider->vpn_service,
+					state, CONNMAN_IPCONFIG_TYPE_IPV6);
 }
 
 int __connman_provider_disconnect(struct connman_provider *provider)
@@ -218,6 +435,8 @@ int __connman_provider_remove(const char *path)
 
 		if (g_strcmp0(srv_path, path) == 0) {
 			DBG("Removing VPN %s", provider->identifier);
+
+			provider_unregister(provider);
 			g_hash_table_remove(provider_hash,
 						provider->identifier);
 			return 0;
@@ -232,7 +451,7 @@ static void provider_append_routes(gpointer key, gpointer value,
 {
 	struct connman_route *route = value;
 	struct connman_provider *provider = user_data;
-	int index = provider->element.index;
+	int index = provider->index;
 
 	if (route->family == AF_INET6) {
 		unsigned char prefix_len = atoi(route->netmask);
@@ -274,6 +493,9 @@ static int set_connected(struct connman_provider *provider,
 		g_hash_table_foreach(provider->routes, provider_append_routes,
 					provider);
 
+		g_hash_table_foreach(provider->user_routes, provider_append_routes,
+					provider);
+
 	} else {
 		if (ipconfig != NULL) {
 			provider_indicate_state(provider,
@@ -342,30 +564,17 @@ int connman_provider_indicate_error(struct connman_provider *provider,
 static void unregister_provider(gpointer data)
 {
 	struct connman_provider *provider = data;
-	struct connman_service *service = provider->vpn_service;
 
-	DBG("provider %p", provider);
+	DBG("provider %p service %p", provider, provider->vpn_service);
 
-	provider->vpn_service = NULL;
-	__connman_service_put(service);
+	if (provider->vpn_service != NULL) {
+		connman_service_unref(provider->vpn_service);
+		provider->vpn_service = NULL;
+	}
 
-	connman_element_unregister(&provider->element);
 	connman_provider_unref(provider);
 }
 
-static void provider_destruct(struct connman_element *element)
-{
-	struct connman_provider *provider = element->private;
-
-	DBG("provider %p", provider);
-
-	g_free(provider->name);
-	g_free(provider->type);
-	g_free(provider->domain);
-	g_free(provider->identifier);
-	g_hash_table_destroy(provider->routes);
-}
-
 static void destroy_route(gpointer user_data)
 {
 	struct connman_route *route = user_data;
@@ -380,17 +589,18 @@ static void provider_initialize(struct connman_provider *provider)
 {
 	DBG("provider %p", provider);
 
-	__connman_element_initialize(&provider->element);
-
-	provider->element.private = provider;
-	provider->element.destruct = provider_destruct;
-
+	provider->index = 0;
 	provider->name = NULL;
 	provider->type = NULL;
 	provider->domain = NULL;
 	provider->identifier = NULL;
+	provider->user_networks = NULL;
 	provider->routes = g_hash_table_new_full(g_direct_hash, g_direct_equal,
 					NULL, destroy_route);
+	provider->user_routes = g_hash_table_new_full(g_str_hash, g_str_equal,
+					g_free, destroy_route);
+	provider->setting_strings = g_hash_table_new_full(g_str_hash, g_str_equal,
+							g_free, g_free);
 }
 
 static struct connman_provider *connman_provider_new(void)
@@ -401,6 +611,8 @@ static struct connman_provider *connman_provider_new(void)
 	if (provider == NULL)
 		return NULL;
 
+	provider->refcount = 1;
+
 	DBG("provider %p", provider);
 	provider_initialize(provider);
 
@@ -425,9 +637,6 @@ static struct connman_provider *connman_provider_get(const char *identifier)
 
 	g_hash_table_insert(provider_hash, provider->identifier, provider);
 
-	provider->element.name = g_strdup(identifier);
-	connman_element_register(&provider->element, NULL);
-
 	return provider;
 }
 
@@ -446,15 +655,143 @@ static void provider_dbus_ident(char *ident)
 	}
 }
 
+static struct connman_provider *provider_create_from_keyfile(GKeyFile *keyfile,
+		const char *ident)
+{
+	struct connman_provider *provider;
+
+	if (keyfile == NULL || ident == NULL)
+		return NULL;
+
+	provider = connman_provider_lookup(ident);
+	if (provider == NULL) {
+		provider = connman_provider_get(ident);
+		if (provider == NULL) {
+			DBG("can not create provider");
+			return NULL;
+		}
+
+		provider_load_from_keyfile(provider, keyfile);
+
+		if (provider->name == NULL || provider->host == NULL ||
+				provider->domain == NULL) {
+			DBG("cannot get name, host or domain");
+			connman_provider_unref(provider);
+			return NULL;
+		}
+
+		provider_register(provider);
+	}
+	return provider;
+}
+
+static int provider_create_service(struct connman_provider *provider)
+{
+	if (provider->vpn_service != NULL)
+		return -EALREADY;
+
+	provider->vpn_service =
+		__connman_service_create_from_provider(provider);
+
+	if (provider->vpn_service == NULL)
+		return -EOPNOTSUPP;
+
+	return 0;
+}
+
+static void provider_create_all_from_type(const char *provider_type)
+{
+	unsigned int i;
+	char **providers;
+	char *id, *type;
+	GKeyFile *keyfile;
+	struct connman_provider *provider;
+
+	DBG("provider type %s", provider_type);
+
+	providers = __connman_storage_get_providers();
+
+	for (i = 0; providers[i] != NULL; i+=1) {
+
+		if (strncmp(providers[i], "provider_", 9) != 0)
+			continue;
+
+		id = providers[i] + 9;
+		keyfile = __connman_storage_load_provider(id);
+
+		if (keyfile == NULL)
+			continue;
+
+		type = g_key_file_get_string(keyfile, id, "Type", NULL);
+
+		DBG("keyfile %p id %s type %s", keyfile, id, type);
+
+		if (strcmp(provider_type, type) != 0) {
+			g_free(type);
+			g_key_file_free(keyfile);
+			continue;
+		}
+
+		provider = provider_create_from_keyfile(keyfile, id);
+		if (provider != NULL) {
+			if (provider_create_service(provider) == -EOPNOTSUPP) {
+				DBG("could not create service");
+				connman_provider_unref(provider);
+			}
+		}
+
+		g_free(type);
+		g_key_file_free(keyfile);
+	}
+	g_strfreev(providers);
+}
+
+static char **get_user_networks(DBusMessageIter *array, int *count)
+{
+	DBusMessageIter entry;
+	char **networks = NULL;
+	GSList *list = NULL, *l;
+	int len;
+
+	dbus_message_iter_recurse(array, &entry);
+
+	while (dbus_message_iter_get_arg_type(&entry) == DBUS_TYPE_STRING) {
+		const char *val;
+		dbus_message_iter_get_basic(&entry, &val);
+
+		list = g_slist_prepend(list, g_strdup(val));
+		dbus_message_iter_next(&entry);
+	}
+
+	len = g_slist_length(list);
+	if (len == 0)
+		goto out;
+
+	networks = g_try_new(char *, len + 1);
+	if (networks == NULL)
+		goto out;
+
+	*count = len;
+	networks[len] = 0;
+
+	for (l = list; l != NULL; l = g_slist_next(l))
+		networks[--len] = l->data;
+
+out:
+	g_slist_free(list);
+
+	return networks;
+}
+
 int __connman_provider_create_and_connect(DBusMessage *msg)
 {
 	struct connman_provider *provider;
 	DBusMessageIter iter, array;
-	const char *type = NULL, *name = NULL, *service_path = NULL;
+	const char *type = NULL, *name = NULL, *service_path;
 	const char *host = NULL, *domain = NULL;
+	char **networks = NULL;
 	char *ident;
-	gboolean created = FALSE;
-	int err;
+	int err, count = 0;
 
 	dbus_message_iter_init(msg, &iter);
 	dbus_message_iter_recurse(&iter, &array);
@@ -480,22 +817,22 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 			else if (g_str_equal(key, "VPN.Domain") == TRUE)
 				dbus_message_iter_get_basic(&value, &domain);
 			break;
+		case DBUS_TYPE_ARRAY:
+			if (g_str_equal(key, "Networks") == TRUE)
+				networks = get_user_networks(&value, &count);
+			break;
 		}
 
 		dbus_message_iter_next(&array);
 	}
 
-	if (host == NULL || domain == NULL) {
-		err = -EINVAL;
-		goto failed;
-	}
+	if (host == NULL || domain == NULL)
+		return -EINVAL;
 
-	DBG("Type %s name %s", type, name);
+	DBG("Type %s name %s networks %p", type, name, networks);
 
-	if (type == NULL || name == NULL) {
-		err = -EOPNOTSUPP;
-		goto failed;
-	}
+	if (type == NULL || name == NULL)
+		return -EOPNOTSUPP;
 
 	ident = g_strdup_printf("%s_%s", host, domain);
 	provider_dbus_ident(ident);
@@ -503,23 +840,30 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 	DBG("ident %s", ident);
 
 	provider = connman_provider_lookup(ident);
-
 	if (provider == NULL) {
-		created = TRUE;
 		provider = connman_provider_get(ident);
-		if (provider) {
-			provider->host = g_strdup(host);
-			provider->domain = g_strdup(domain);
-			provider->name = g_strdup(name);
-			provider->type = g_strdup(type);
+		if (provider == NULL) {
+			DBG("can not create provider");
+			g_free(ident);
+			return -EOPNOTSUPP;
 		}
+
+		provider->host = g_strdup(host);
+		provider->domain = g_strdup(domain);
+		provider->name = g_strdup(name);
+		provider->type = g_strdup(type);
+
+		if (provider_register(provider) == 0)
+			connman_provider_load(provider);
 	}
 
-	if (provider == NULL) {
-		DBG("can not create provider");
-		err = -EOPNOTSUPP;
-		goto failed;
+	if (networks != NULL) {
+		g_strfreev(provider->user_networks);
+		provider->user_networks = networks;
+		provider->num_user_networks = count;
+		set_user_networks(provider, provider->user_networks);
 	}
+
 	dbus_message_iter_init(msg, &iter);
 	dbus_message_iter_recurse(&iter, &array);
 
@@ -545,26 +889,21 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 
 	g_free(ident);
 
-	if (provider == NULL) {
-		err = -EOPNOTSUPP;
-		goto failed;
+	err = provider_create_service(provider);
+	if (err == -EALREADY) {
+		DBG("provider already connected");
+	} else {
+		if (err == -EOPNOTSUPP) {
+			goto unref;
+		} else {
+			err = __connman_service_connect(provider->vpn_service);
+
+			if (err < 0 && err != -EINPROGRESS)
+				goto failed;
+		}
 	}
 
-	if (created == TRUE)
-		provider_probe(provider);
-
-	if (provider->vpn_service == NULL)
-		provider->vpn_service =
-			__connman_service_create_from_provider(provider);
-	if (provider->vpn_service == NULL) {
-		err = -EOPNOTSUPP;
-		goto failed;
-	}
-
-	err = __connman_service_connect(provider->vpn_service);
-	if (err < 0 && err != -EINPROGRESS)
-		goto failed;
-
+	connman_provider_save(provider);
 	service_path = __connman_service_get_path(provider->vpn_service);
 	g_dbus_send_reply(connection, msg,
 				DBUS_TYPE_OBJECT_PATH, &service_path,
@@ -572,15 +911,13 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 	return 0;
 
 failed:
-	if (provider != NULL && created == TRUE) {
-		DBG("can not connect delete provider");
-		connman_provider_unref(provider);
+	connman_service_unref(provider->vpn_service);
+	provider->vpn_service = NULL;
 
-		if (provider->vpn_service != NULL) {
-			__connman_service_put(provider->vpn_service);
-			provider->vpn_service = NULL;
-		}
-	}
+unref:
+	DBG("can not connect, delete provider");
+
+	g_hash_table_remove(provider_hash, provider->identifier);
 
 	return err;
 }
@@ -604,9 +941,16 @@ int connman_provider_set_string(struct connman_provider *provider,
 	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(provider->name);
 		provider->name = g_strdup(value);
-	}
-
-	return connman_element_set_string(&provider->element, key, value);
+	} else if (g_str_equal(key, "Host") == TRUE) {
+		g_free(provider->host);
+		provider->host = g_strdup(value);
+	} else if (g_str_equal(key, "VPN.Domain") == TRUE) {
+		g_free(provider->domain);
+		provider->domain = g_strdup(value);
+	} else
+		g_hash_table_replace(provider->setting_strings,
+				g_strdup(key), g_strdup(value));
+	return 0;
 }
 
 const char *connman_provider_get_string(struct connman_provider *provider,
@@ -618,8 +962,29 @@ const char *connman_provider_get_string(struct connman_provider *provider,
 		return provider->type;
 	else if (g_str_equal(key, "Name") == TRUE)
 		return provider->name;
+	else if (g_str_equal(key, "Host") == TRUE)
+		return provider->host;
+	else if (g_str_equal(key, "VPN.Domain") == TRUE)
+		return provider->domain;
 
-	return connman_element_get_string(&provider->element, key);
+	return g_hash_table_lookup(provider->setting_strings, key);
+}
+
+connman_bool_t
+__connman_provider_check_routes(struct connman_provider *provider)
+{
+	if (provider == NULL)
+		return FALSE;
+
+	if (provider->user_routes != NULL &&
+			g_hash_table_size(provider->user_routes) > 0)
+		return TRUE;
+
+	if (provider->routes != NULL &&
+			g_hash_table_size(provider->routes) > 0)
+		return TRUE;
+
+	return FALSE;
 }
 
 void *connman_provider_get_data(struct connman_provider *provider)
@@ -654,7 +1019,7 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 		}
 	}
 
-	connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_FIXED);
+	__connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_FIXED);
 	__connman_ipconfig_set_index(ipconfig, index);
 
 
@@ -670,16 +1035,16 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 		}
 	}
 
-	connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_OFF);
+	__connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_OFF);
 	__connman_ipconfig_set_index(ipconfig, index);
 
 done:
-	provider->element.index = index;
+	provider->index = index;
 }
 
 int connman_provider_get_index(struct connman_provider *provider)
 {
-	return provider->element.index;
+	return provider->index;
 }
 
 int connman_provider_set_ipaddress(struct connman_provider *provider,
@@ -743,7 +1108,7 @@ int connman_provider_set_nameservers(struct connman_provider *provider,
 
 	for (i = 0; nameservers_array[i] != NULL; i++) {
 		__connman_service_nameserver_append(provider->vpn_service,
-							nameservers_array[i]);
+						nameservers_array[i], FALSE);
 	}
 
 	g_strfreev(nameservers_array);
@@ -858,9 +1223,17 @@ int connman_provider_append_route(struct connman_provider *provider,
 
 const char *connman_provider_get_driver_name(struct connman_provider *provider)
 {
+	if (provider->driver == NULL)
+		return NULL;
+
 	return provider->driver->name;
 }
 
+const char *connman_provider_get_save_group(struct connman_provider *provider)
+{
+	return provider->identifier;
+}
+
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
 	return 0;
@@ -880,6 +1253,7 @@ int connman_provider_driver_register(struct connman_provider_driver *driver)
 
 	driver_list = g_slist_insert_sorted(driver_list, driver,
 							compare_priority);
+	provider_create_all_from_type(driver->name);
 	return 0;
 }
 
@@ -890,12 +1264,12 @@ void connman_provider_driver_unregister(struct connman_provider_driver *driver)
 	driver_list = g_slist_remove(driver_list, driver);
 }
 
-static void provider_remove(gpointer key, gpointer value,
+static void provider_remove_all(gpointer key, gpointer value,
 						gpointer user_data)
 {
 	struct connman_provider *provider = value;
 
-	g_hash_table_remove(provider_hash, provider->identifier);
+	__connman_provider_remove(provider->identifier);
 }
 
 static void provider_offline_mode(connman_bool_t enabled)
@@ -903,13 +1277,75 @@ static void provider_offline_mode(connman_bool_t enabled)
 	DBG("enabled %d", enabled);
 
 	if (enabled == TRUE)
-		g_hash_table_foreach(provider_hash, provider_remove, NULL);
+		g_hash_table_foreach(provider_hash, provider_remove_all, NULL);
 
 }
 
+static struct connman_provider *provider_get(int index)
+{
+	GHashTableIter iter;
+	gpointer value, key;
+
+	g_hash_table_iter_init(&iter, provider_hash);
+
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		struct connman_provider *provider = value;
+
+		if (provider->index == index)
+			return provider;
+	}
+
+	return NULL;
+}
+
+static void provider_service_changed(struct connman_service *service,
+				enum connman_service_state state)
+{
+	struct connman_provider *provider;
+	int vpn_index, service_index;
+
+	if (service == NULL)
+		return;
+
+	switch (state) {
+	case CONNMAN_SERVICE_STATE_UNKNOWN:
+	case CONNMAN_SERVICE_STATE_IDLE:
+	case CONNMAN_SERVICE_STATE_ASSOCIATION:
+	case CONNMAN_SERVICE_STATE_CONFIGURATION:
+	case CONNMAN_SERVICE_STATE_READY:
+	case CONNMAN_SERVICE_STATE_ONLINE:
+		return;
+	case CONNMAN_SERVICE_STATE_DISCONNECT:
+	case CONNMAN_SERVICE_STATE_FAILURE:
+		break;
+	}
+
+	service_index = __connman_service_get_index(service);
+
+	vpn_index = __connman_connection_get_vpn_index(service_index);
+
+	DBG("service %p %s state %d index %d/%d", service,
+		__connman_service_get_ident(service),
+		state, service_index, vpn_index);
+
+	if (vpn_index < 0)
+		return;
+
+	provider = provider_get(vpn_index);
+	if (provider == NULL)
+		return;
+
+	DBG("disconnect %p index %d", provider, vpn_index);
+
+	__connman_provider_disconnect(provider);
+
+	return;
+}
+
 static struct connman_notifier provider_notifier = {
 	.name			= "provider",
 	.offline_mode		= provider_offline_mode,
+	.service_state_changed	= provider_service_changed,
 };
 
 int __connman_provider_init(void)
